import pandas as pd


def alifestd_make_empty() -> pd.DataFrame:
    """Create an alife standard phylogeny dataframe with zero rows."""
    return pd.DataFrame(
        {
            "ancestor_list": pd.Series(dtype="str"),
            "id": pd.Series(dtype="int"),
        }
    )



import typing

from .._auxiliary_lib import get_hstrat_version
from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._pack_differentiae import pack_differentiae


def col_to_records(column: HereditaryStratigraphicColumn) -> typing.Dict:
    """Serialize a `HereditaryStratigraphicColumn` to a dict composed of
    builtin types."""
    differentia_bit_width = column.GetStratumDifferentiaBitWidth()
    packed_differentiae = pack_differentiae(
        column.IterRetainedStrata(),
        differentia_bit_width,
    )

    policy = column._stratum_retention_policy
    spec = policy.GetSpec()
    res = {
        "policy_algo": spec.GetAlgoIdentifier(),
        "policy_spec": {
            k.lstrip("_"): v  # strip leading underscores on private members
            for k, v in spec.__dict__.items()
        },
        "policy": policy.GetEvalCtor(),
        "num_strata_deposited": column.GetNumStrataDeposited(),
        "differentiae": packed_differentiae,
        "differentia_bit_width": differentia_bit_width,
        "hstrat_version": get_hstrat_version(),
    }

    if not column._ShouldOmitStratumDepositionRank():
        res["stratum_deposition_ranks"] = [
            stratum.GetDepositionRank()
            for stratum in column.IterRetainedStrata()
        ]

    if column.HasAnyAnnotations():
        res["stratum_annotations"] = [
            stratum.GetAnnotation() for stratum in column.IterRetainedStrata()
        ]

    return res



from base64 import b64decode
import typing

from bitstring import BitArray
import numpy as np

from .._auxiliary_lib import iter_chunks


def _numpy_unpack_differentiae(
    _bytes: bytes, differentia_bit_width: int
) -> typing.Iterator[int]:
    # numpy dtypes can be composed of various attributes
    # in this case, we want our type to be big-endian (>),
    # and to be composed of `num_bytes` unsigned ints (u).
    # for more info, see https://numpy.org/doc/stable/reference/generated/numpy.dtype.html
    num_bytes = differentia_bit_width // 8
    dt = np.dtype(f">u{num_bytes}")
    yield from np.frombuffer(_bytes, dtype=dt)


def _bitarray_unpack_differentiae(
    _bytes: bytes, differentia_bit_width: int
) -> typing.Iterator[int]:
    bits = BitArray(bytes=_bytes)
    num_padding_bits = bits[:8].uint
    if num_padding_bits:
        valid_bits = bits[8:-num_padding_bits]
    else:
        valid_bits = bits[8:]

    assert len(valid_bits) % differentia_bit_width == 0

    for chunk in iter_chunks(valid_bits, differentia_bit_width):
        yield chunk.uint


def unpack_differentiae(
    packed_differentiae: str,
    differentia_bit_width: int,
) -> typing.Iterator[int]:
    """Unpack a compact, concatenated base 64 representation into
    a sequence with each element represented as a distinct integer."""
    _bytes = b64decode(packed_differentiae)

    if differentia_bit_width in [8, 16, 32, 64]:
        # bit width is a multiple of 8 -- no padding required
        yield from _numpy_unpack_differentiae(_bytes, differentia_bit_width)
    else:
        # padding bits are possible, first byte tells how many are required
        yield from _bitarray_unpack_differentiae(_bytes, differentia_bit_width)



from base64 import b64encode
import sys
import typing

from bitarray import util as bitarray_util
import numpy as np

from .._auxiliary_lib import iter_slices, zip_strict
from ..genome_instrumentation import HereditaryStratum


def _make_buffer_bitarray(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> typing.ByteString:
    try:
        len(strata)
    except TypeError:
        strata = [*strata]

    needs_header = bool(differentia_bit_width % 8)
    buffer_size = 8 * needs_header + len(strata) * differentia_bit_width
    buffer = bitarray_util.zeros(buffer_size)

    # use first byte to tell how many padding bits if any are required
    if needs_header:
        num_jagged_bits = len(buffer) % 8
        num_padding_bits = (8 - num_jagged_bits) % 8
        header = bitarray_util.int2ba(
            num_padding_bits,
            length=8,
        )
        buffer[:8] |= header

    for slice_, stratum in zip_strict(
        iter_slices(len(buffer), differentia_bit_width, needs_header * 8),
        strata,
    ):
        buffer[slice_] |= bitarray_util.int2ba(
            stratum.GetDifferentia(), length=differentia_bit_width
        )

    return buffer


def _make_buffer_numpy(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> typing.ByteString:
    buffer = np.fromiter(
        (s.GetDifferentia() for s in strata),
        dtype=eval(f"np.uint{differentia_bit_width}"),
    )
    if {
        ">": "big",
        "<": "little",
        "=": sys.byteorder,
        "|": "not applicable",
    }[buffer.dtype.byteorder] == "little":
        buffer.byteswap(inplace=True)
    return buffer


def _make_buffer(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> typing.ByteString:
    strata = [*strata]
    if differentia_bit_width in (8, 16, 32, 64):
        return _make_buffer_numpy(strata, differentia_bit_width)
    else:
        return _make_buffer_bitarray(strata, differentia_bit_width)


def pack_differentiae(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> str:
    """Pack a sequence of differentiae together into a compact
    representation.

    Returns a string with base 64 encoded concatenation of diffferentiae.
    If `differentia_bit_width` is not an even byte multiple, the first encoded
    byte tells how many empty padding bits, if any, were placed at the end of
    the concatenation in order to align the bitstring end to byte boundaries.
    """

    buffer = _make_buffer(strata, differentia_bit_width)

    encoded_bytes = b64encode(buffer.tobytes())

    encoded_str = encoded_bytes.decode()
    return encoded_str



def calc_provided_uncertainty(
    depth_proportional_resolution: int,
    num_stratum_depositions_completed: int,
) -> int:
    """After n strata have been deposited, how many ranks are spaced
    between retained strata?

    Note that the returned value actually corresponds to one more than the
    uncertainty with respect to calculating MRCA supposing two identically
    distributed columns. For example a return value of 1 corresponds to
    strata retained at every rank, so the rank of the MRCA can be
    determined with 0 uncertainty.
    """
    guaranteed_resolution = depth_proportional_resolution
    max_uncertainty = (
        num_stratum_depositions_completed // guaranteed_resolution
    )

    # round down to lower or equal power of 2
    # cast to int to make robust to numpy.int32, numpy.int64, etc.
    provided_uncertainty_exp = int(max_uncertainty // 2).bit_length()
    provided_uncertainty = 2**provided_uncertainty_exp
    return provided_uncertainty



"""Implementation helpers."""

from ._col_records_from_pop_records import col_records_from_pop_records
from ._policy_from_record import policy_from_record

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "col_records_from_pop_records",
    "policy_from_record",
]



def calc_provided_uncertainty(
    depth_proportional_resolution: int,
    num_stratum_depositions_completed: int,
) -> int:
    """After n strata have been deposited, how many ranks are spaced between
    retained strata?

    Note that the returned value actually corresponds to one more than the
    uncertainty with respect to calculating MRCA supposing two identically
    distributed columns. For example a return value of 1 corresponds to strata
    retained at every rank, so the rank of the MRCA can be determined with 0
    uncertainty.
    """
    guaranteed_resolution = depth_proportional_resolution
    max_uncertainty = (
        num_stratum_depositions_completed // guaranteed_resolution
    )

    # round down to lower or equal power of 2
    # cast to int to make robust to numpy.int32, numpy.int64, etc.
    provided_uncertainty_exp = int(max_uncertainty // 2).bit_length()
    provided_uncertainty = 2**provided_uncertainty_exp
    return provided_uncertainty
